@@ -1,6 +1,6 @@
 /***************************************************************************
  begin       : Wed Apr 28 2010
- copyright   : (C) 2010 by Martin Preuss
+ copyright   : (C) 2010, 2016 by Martin Preuss
  email       : martin@libchipcard.de
 
  ***************************************************************************
@@ -264,7 +264,8 @@ int GWEN_SyncIo_Tls__readFile(const char *fname, GWEN_BUFFER *buf) {
 
 
 
-#if 0
+#if GWEN_TLS_USE_SYSTEM_CERTIFICATES
+# ifndef OS_WIN32
 static int GWEN_SyncIo_Tls_AddCaCertFolder(GWEN_SYNCIO *sio, const char *folder) {
   GWEN_SYNCIO_TLS *xio;
   int rv;
@@ -294,16 +295,15 @@ static int GWEN_SyncIo_Tls_AddCaCertFolder(GWEN_SYNCIO *sio, const char *folder)
         s=GWEN_StringListEntry_Data(se);
         if (s && *s) {
           rv=gnutls_certificate_set_x509_trust_file(xio->credentials,
-              s,
-              GNUTLS_X509_FMT_PEM);
+						    s,
+                                                    GNUTLS_X509_FMT_PEM);
           if (rv<=0) {
             DBG_WARN(GWEN_LOGDOMAIN,
                      "gnutls_certificate_set_x509_trust_file(%s): %d (%s)",
                      s, rv, gnutls_strerror(rv));
           }
           else {
-            DBG_INFO(GWEN_LOGDOMAIN,
-                     "Added %d trusted certs from [%s]", rv, s);
+            DBG_INFO(GWEN_LOGDOMAIN, "Added %d trusted certs from [%s]", rv, s);
             successfullTustFileCount++;
           }
         }
@@ -320,6 +320,7 @@ static int GWEN_SyncIo_Tls_AddCaCertFolder(GWEN_SYNCIO *sio, const char *folder)
 
   return successfullTustFileCount;
 }
+# endif
 #endif
 
 
@@ -408,8 +409,16 @@ int GWEN_SyncIo_Tls_Prepare(GWEN_SYNCIO *sio) {
 
   /* find default trust file if none is selected */
   if (lflags & GWEN_SYNCIO_TLS_FLAGS_ADD_TRUSTED_CAS) {
+#if GWEN_TLS_USE_SYSTEM_CERTIFICATES
+    /* disable setting of default trust file as discussed on aqbanking-users.
+     * The rationale is that without this file being set gnutls should behave
+     * correctly on each system.
+     * On Linux systems it should use the standard mechanism of the underlying
+     * distribution. On Windows the default CA store should be used (if given
+     * "--with-default-trust-store-file" to "./configure" of GNUTLS).
+     */
     int trustFileSet=0;
-#if 0
+
 # ifndef OS_WIN32
     /* try to find OpenSSL certificates */
     if (trustFileSet==0) {
@@ -420,9 +429,7 @@ int GWEN_SyncIo_Tls_Prepare(GWEN_SYNCIO *sio) {
       GWEN_StringList_AppendString(paths, "/etc/ssl/certs", 0, 0);
 
       nbuf=GWEN_Buffer_new(0, 256, 0, 1);
-      rv=GWEN_Directory_FindFileInPaths(paths,
-                                        "ca-certificates.crt",
-                                        nbuf);
+      rv=GWEN_Directory_FindFileInPaths(paths, "ca-certificates.crt", nbuf);
       GWEN_StringList_free(paths);
       if (rv==0) {
         DBG_INFO(GWEN_LOGDOMAIN,
@@ -464,47 +471,20 @@ int GWEN_SyncIo_Tls_Prepare(GWEN_SYNCIO *sio) {
     }
 
 # endif
-#endif
 
 
     if (trustFileSet==0) {
-      GWEN_STRINGLIST *paths;
 
-      /* try to find our trust file */
-      paths=GWEN_PathManager_GetPaths(GWEN_PM_LIBNAME, GWEN_PM_DATADIR);
-      if (paths) {
-        GWEN_BUFFER *nbuf;
-
-        nbuf=GWEN_Buffer_new(0, 256, 0, 1);
-        rv=GWEN_Directory_FindFileInPaths(paths,
-                                          "ca-bundle.crt",
-                                          nbuf);
-        GWEN_StringList_free(paths);
-        if (rv==0) {
-          DBG_INFO(GWEN_LOGDOMAIN,
-                   "Using default ca-bundle from [%s]",
-                   GWEN_Buffer_GetStart(nbuf));
-          rv=gnutls_certificate_set_x509_trust_file(xio->credentials,
-              GWEN_Buffer_GetStart(nbuf),
-              GNUTLS_X509_FMT_PEM);
-          if (rv<=0) {
-            DBG_ERROR(GWEN_LOGDOMAIN,
-                      "gnutls_certificate_set_x509_trust_file(%s): %d (%s)",
-                      GWEN_Buffer_GetStart(nbuf), rv, gnutls_strerror(rv));
-          }
-          else {
-            DBG_INFO(GWEN_LOGDOMAIN,
-                     "Added %d trusted certs", rv);
-            trustFileSet=1;
-          }
-        }
-        GWEN_Buffer_free(nbuf);
-      }
+        /* TODO: use gnutls_certificate_set_x509_system_trust() */
+        trustFileSet=1;
     }
 
+
+
     if (trustFileSet==0) {
       DBG_WARN(GWEN_LOGDOMAIN, "No default bundle file found");
     }
+#endif
   }
 
   /* possibly set trust file */
@@ -541,8 +521,8 @@ int GWEN_SyncIo_Tls_Prepare(GWEN_SYNCIO *sio) {
       return rv;
     }
     else {
-      gnutls_datum d;
-      gnutls_dh_params dh_params=NULL;
+      gnutls_datum_t d;
+      gnutls_dh_params_t dh_params=NULL;
 
       rv=gnutls_dh_params_init(&dh_params);
       if (rv<0) {
@@ -753,11 +733,11 @@ int GWEN_SyncIo_Tls_GetPeerCert(GWEN_SYNCIO *sio) {
           gcry_free(e.data);
       }
 
-      /* get fingerprint */
+      /* get fingerprint (MD5) */
       size=16;
       rv=gnutls_x509_crt_get_fingerprint(cert, GNUTLS_DIG_MD5, buffer1, &size);
       if (rv!=0) {
-        DBG_INFO(GWEN_LOGDOMAIN, "gnutls_x509_crt_get_fingerprint: %d (%s)", rv, gnutls_strerror(rv));
+        DBG_INFO(GWEN_LOGDOMAIN, "gnutls_x509_crt_get_fingerprint(MD5): %d (%s)", rv, gnutls_strerror(rv));
         GWEN_SslCertDescr_free(certDescr);
         gnutls_x509_crt_deinit(cert);
         return GWEN_ERROR_GENERIC;
@@ -777,6 +757,55 @@ int GWEN_SyncIo_Tls_GetPeerCert(GWEN_SYNCIO *sio) {
         GWEN_Buffer_free(dbuf);
       }
 
+      /* get fingerprint (SHA1) */
+      size=sizeof(buffer1);
+      rv=gnutls_x509_crt_get_fingerprint(cert, GNUTLS_DIG_SHA1, buffer1, &size);
+      if (rv!=0) {
+        DBG_INFO(GWEN_LOGDOMAIN, "gnutls_x509_crt_get_fingerprint(SHA1): %d (%s)", rv, gnutls_strerror(rv));
+        GWEN_SslCertDescr_free(certDescr);
+        gnutls_x509_crt_deinit(cert);
+        return GWEN_ERROR_GENERIC;
+      }
+      else {
+        GWEN_BUFFER *dbuf;
+
+        dbuf=GWEN_Buffer_new(0, 256, 0, 1);
+        if (GWEN_Text_ToHexBuffer(/* GCC4 pointer-signedness fix: */ buffer1,
+            size, dbuf, 2, ':', 0)) {
+          DBG_ERROR(GWEN_LOGDOMAIN,
+                    "Could not convert fingerprint to hex");
+        }
+        else {
+          GWEN_SslCertDescr_SetFingerPrintSha1(certDescr, GWEN_Buffer_GetStart(dbuf));
+        }
+        GWEN_Buffer_free(dbuf);
+      }
+
+      /* get fingerprint (SHA512) */
+      size=sizeof(buffer1);
+      rv=gnutls_x509_crt_get_fingerprint(cert, GNUTLS_DIG_SHA512, buffer1, &size);
+      if (rv!=0) {
+        DBG_INFO(GWEN_LOGDOMAIN, "gnutls_x509_crt_get_fingerprint(SHA512): %d (%s)", rv, gnutls_strerror(rv));
+        GWEN_SslCertDescr_free(certDescr);
+        gnutls_x509_crt_deinit(cert);
+        return GWEN_ERROR_GENERIC;
+      }
+      else {
+        GWEN_BUFFER *dbuf;
+
+        dbuf=GWEN_Buffer_new(0, 256, 0, 1);
+        if (GWEN_Text_ToHexBuffer(/* GCC4 pointer-signedness fix: */ buffer1,
+            size, dbuf, 2, ':', 0)) {
+          DBG_ERROR(GWEN_LOGDOMAIN,
+                    "Could not convert fingerprint to hex");
+        }
+        else {
+          GWEN_SslCertDescr_SetFingerPrintSha512(certDescr, GWEN_Buffer_GetStart(dbuf));
+        }
+        GWEN_Buffer_free(dbuf);
+      }
+
+
       if (xio->hostName) {
         DBG_INFO(GWEN_LOGDOMAIN, "Checking hostname [%s]", xio->hostName);
         if (!gnutls_x509_crt_check_hostname(cert, xio->hostName)) {
@@ -1115,7 +1144,9 @@ void GWEN_SyncIo_Tls_ShowCipherInfo(GWEN_SYNCIO *sio) {
   case GNUTLS_CIPHER_CAMELLIA_256_CBC:
   case GNUTLS_CIPHER_RC2_40_CBC:
   case GNUTLS_CIPHER_DES_CBC:
-  case GNUTLS_CIPHER_AES_192_CBC: // new in gnutls-2.9.8, so i.e. not available in gnutls-2.8.x
+#ifdef GNUTLS_CIPHER_AES_192_CBC
+  case GNUTLS_CIPHER_AES_192_CBC: /* new in gnutls-2.9.8, so i.e. not available in gnutls-2.8.x */
+#endif
   default:
     break;
   }
@@ -1315,7 +1346,14 @@ int GWENHYWFAR_CB GWEN_SyncIo_Tls_Read(GWEN_SYNCIO *sio,
     GWEN_SyncIo_Disconnect(baseIo);
 #ifdef GNUTLS_E_PREMATURE_TERMINATION
     if (rv==GNUTLS_E_PREMATURE_TERMINATION) {
-      return GWEN_ERROR_SSL_PREMATURE_CLOSE;
+      if (GWEN_SyncIo_GetFlags(sio) & GWEN_SYNCIO_TLS_FLAGS_IGN_PREMATURE_CLOSE) {
+	DBG_ERROR(GWEN_LOGDOMAIN, "Detected premature disconnect by server (violates specs!), ignoring.");
+	return 0; /* report EOF */
+      }
+      else {
+	DBG_ERROR(GWEN_LOGDOMAIN, "Detected premature disconnect by server (violates specs!)");
+	return GWEN_ERROR_SSL_PREMATURE_CLOSE;
+      }
     }
 #endif
     return GWEN_ERROR_SSL;
